/*
 * xtags - print C-file's indexes
 *
 * Program zaliczeniowy
 * Autor: Micha Moskal (nr albumu 129934)
 *
 * $Id: parser.c,v 1.3 2000/10/24 22:52:12 malekith Exp $
 */

#include "xtags.h"

/* possible states of parser */
enum { 
	s_norm = 1, 	/* regular */
	s_string, 	/* inside "" '' */
	s_comment, 	/* inside |* *| =~ s/|/\// */
	s_comment1, 	/* inside // \n */
	s_cpp 		/* inside # (\\\n)+ \n */
};

struct token {
	struct token *left, *right;
	int type;
	char *val;
	int def;
	
	/* where it was info */
	int *lines;
	int lines_size, lines_ptr;
};

struct parser_state {
	FILE *in;
	const char *filename;
	int state;
	int close_delim;	/* second part of state, '"', '\'' */
	int level;		/* lev of {}, 0 for file scope */
	struct token *ttree;
	int line_no;

	char *line;
	int llen;

	char *word;
	int word_line;
	int word_lev;
	/* `call' is used to detect function calls and defintions 
	 * 0 - after id, 1 - after (, -1 after anything else
	 * comments and white space don't change it, so
	 * 	func_name |* comment *| ();
	 * will be trated correctly
	 */
	int call;		
};

#define CTX struct parser_state

static int get_line(CTX *ctx);
static void get_id(CTX *ctx, char **pp);
static void add_line(struct token *t, int l);
static struct token *add_word(CTX *ctx, struct token *t);
static void flush_word(CTX *ctx);
static char *norm_char(CTX *ctx, char *p);
static void cpp(CTX *ctx, char *p);
static void handle_line(CTX *ctx);
static void dump_token(struct token *t, FILE *out);
static void dump(CTX *ctx, FILE *out);

/*
 * get line from ctx->in into dynamicly enlarged buffer
 * ret -1 in case of EOF, 0 == success
 */
static int get_line(CTX *ctx)
{
	int i, c;

	for (i = 0; ; i++) {
		if (i >= ctx->llen)
			ctx->line = (char*)realloc(ctx->line, 
						   ctx->llen *= 2);
		c = fgetc(ctx->in);

		/* skip those ugly M$ CRs */
		if (c == '\r') {
			i--;
			continue;
		}

		/* check for preprocessor \ continued lines */
		if (c == '\n' && i)
			if (ctx->line[i - 1] == '\\') {
				i -= 2;		/* and join them */
				ctx->line_no++;
				continue;
			}
			
		if (c == EOF || c == '\n')
			break;
			
		ctx->line[i] = c;
	}

	ctx->line[i] = 0;

	ctx->line_no++;
	
	if (verbose > 1)
		fprintf(stderr, "%s:%d: %s\n", 
			ctx->filename, ctx->line_no, ctx->line);

	return -(i == 0 && c == EOF);
}

static void get_id(CTX *ctx, char **pp)
{
	char *p, *s;

	s = p = *pp;
	while (isalnum(*p) || *p == '_')
		p++;
	ctx->word = (char*)malloc(p - s + 1);
	memcpy(ctx->word, s, p - s);
	ctx->word[p - s] = 0;
	ctx->word_line = ctx->line_no;
	ctx->word_lev = ctx->level;
	*pp = p;
}

static void add_line(struct token *t, int l)
{
	if (t->lines_ptr == t->lines_size)
		t->lines = (int*)realloc(t->lines, 
				sizeof(int) * (t->lines_size *= 2));
	t->lines[t->lines_ptr++] = l;
}

static struct token *add_word(CTX *ctx, struct token *t)
{
	int d;

	if (t == 0) {
		t = (struct token*)malloc(sizeof(struct token));
		t->left = t->right = 0;
		t->type =
			ctx->call == -2 ?
				t_cpp :
			is_keyword(ctx->word) ?
				t_keyword : 
			ctx->call == 1 ? 
				t_func : 
				t_global;
				
		t->val = ctx->word;
		ctx->word = 0;

		t->lines = (int*)malloc(sizeof(int) * (t->lines_size = 16));
		t->lines_ptr = 0;
		t->def = 0;

		if (t->type == t_cpp || ctx->word_lev == 0)
			t->def = ctx->word_line;
		else {
			add_line(t, ctx->word_line);
			if (t->type == t_global)
				t->type = t_local;
		}

		return t;
	}
	
	d = strcmp(t->val, ctx->word);

	if (d > 0)
		t->left = add_word(ctx, t->left);
	else if (d < 0)
		t->right = add_word(ctx, t->right);
	else {
		if (ctx->word_lev == 0) {
			if (t->type == t_func ||
			    t->type == t_local) {
				if (t->def)
					add_line(t, t->def);
				if (t->type == t_local)
					t->type = t_global;
				t->def = ctx->word_line;
			} else if (t->def == 0)
				t->def = ctx->word_line;
			else
				add_line(t, ctx->word_line);
		} else
			add_line(t, ctx->word_line);
	}

	return t;
}

static void flush_word(CTX *ctx)
{
	if (ctx->word == 0 || ctx->call == 0)
		return;

	if (verbose > 2)
		fprintf(stderr, "\tword: %s%s\n", 
			ctx->word, ctx->call == 1 ? "()" : "");

	ctx->ttree = add_word(ctx, ctx->ttree);
	
	free(ctx->word);
	ctx->word = 0;
}

static char *norm_char(CTX *ctx, char *p)
{
	switch (*p) {
	case '/':
		if (allow_plus_comments && p[1] == '/')
			ctx->state = s_comment1;
		else if (p[1] == '*')
			ctx->state = s_comment;
		else
			ctx->call = -1;
		break;
	case '{':
		ctx->level++;
		ctx->call = -1;
		break;
	case '}':
		ctx->level--;
		ctx->call = -1;
		break;
	case '\'':
	case '"':
		ctx->state = s_string;
		ctx->close_delim = *p;
		ctx->call = -1;
		break;
	case '(':
		ctx->call = 1;
		break;
	case '0':
		ctx->call = -1;
		if (tolower(p[1]) == 'x') {
			p += 2;
			while (isxdigit(*p))
				p++;
		} else {
			p++;
			while (*p >= '0' && *p <= '7')
				p++;
		}
		p--;
		break;
	default:
		if (isdigit(*p)) {
			p++;
			while (isdigit(*p))
				p++;
			p--;
			ctx->call = -1;
		} else if (isalpha(*p) || *p == '_') {
			ctx->call = -1;
			flush_word(ctx);
			ctx->call = 0;
			get_id(ctx, &p);
			p--;
		} else if (isspace(*p)) {
			/* nothing */;
		} else
			ctx->call = -1;
		break;
	}

	return p;
}

/*
 * parse preprocessor lines... 
 * only #define at the moment is supported
 * note that it isn't perfect, esp:
 *	#define |* comment *| sth XXX
 * can fool it
 */
static void cpp(CTX *ctx, char *p)
{
	ctx->call = -1;
	flush_word(ctx);

	while (isspace(*p))
		p++;
		
	if (strncmp(p, "define", 6) == 0) {
		p += 6;
		while (isspace(*p))
			p++;

		if (isalpha(*p)) {
			get_id(ctx, &p);
			ctx->call = -2;
			flush_word(ctx);
		}
	}
}

static void handle_line(CTX *ctx)
{
	char *p;

	p = ctx->line;
	
	while (isspace(*p))
		p++;
		
	if (ctx->state == s_norm && *p == '#') {
		cpp(ctx, p + 1);
		return;
	}

	for (; *p; p++) {
		switch (ctx->state) {
		case s_norm:
			p = norm_char(ctx, p);
			break;
		case s_comment:
			if (p[0] == '*' && p[1] == '/') {
				ctx->state = s_norm;
				p++;
			}
			break;
		case s_comment1:
			/* nothing to be done till end of the line */
			break;
		case s_string:
			if (*p == '\\' && p[1])
				p++;
			else if (*p == ctx->close_delim)
				ctx->state = s_norm;
			break;
		default:
			abort();
		}
		flush_word(ctx);
	}

	/* comment1 (//) cannot cross lines */
	if (ctx->state == s_comment1)
		ctx->state = s_norm;
}

static void dump_token(struct token *t, FILE *out)
{
	int i;
	
	if (t == 0) 
		return;
		
	dump_token(t->left, out);

	if (t->type & print_flags) {
		fprintf(out, "\t- %s%s [%d]: ", t->val, 
			t->type == t_func ? 
				"()" : 
			t->type == t_cpp ?
				"#" :
				"", 
			t->def);

		for (i = 0; i < t->lines_ptr; i++)
			fprintf(out, "%d%s", 
				t->lines[i], 
				i + 1 == t->lines_ptr ?  "." : ", ");
		fprintf(out, "\n");
	}

	dump_token(t->right, out);

	free(t->lines);
	free(t->val);
	free(t);
}

static void dump(CTX *ctx, FILE *out)
{
	if (verbose)
		fprintf(stderr, "dumping %s...\n", ctx->filename);
	fprintf(out, "Identifiers of %s:\n", ctx->filename);
	dump_token(ctx->ttree, out);
}

/*
 * the main (and only ;) entry point for this file ...
 * reads `in', writes results to `out'
 */
void parse(FILE *in, FILE *out, const char *name)
{
	CTX ctx;
	
	if (verbose)
		fprintf(stderr, "processing %s...\n", name);

	ctx.in = in;
	ctx.state = s_norm;
	ctx.filename = name;
	ctx.level = 0;
	ctx.ttree = 0;
	ctx.line = (char*)malloc(ctx.llen = 80);
	ctx.line_no = 0;
	ctx.word = 0;

	while (get_line(&ctx) == 0)
		handle_line(&ctx);

	ctx.call = -1;
	flush_word(&ctx);
	
	dump(&ctx, out);
	free(ctx.line);
}
